from DiscordDatabase.common_functions import format_string
from DiscordDatabase.database import Database


class DiscordDatabase:
    # This class takes care of creating categories and channels for the database
    def __init__(self, client, guild_id) -> None:
        self.guild_id = guild_id
        self.client = client

    async def __create(self, category_name: str, channel_name: str):
        category_name = format_string(category_name)  # No spaces allowed
        channel_name = format_string(channel_name)  # No spaces allowed

        if len(category_name) <= 0:
            raise ValueError("category_name should atleast have a length of 1")
        if len(channel_name) <= 0:
            raise ValueError("channel_name should atleast have a length of 1")

        ##### CATEGORY #####
        category = list(
            filter(lambda c: c.name.casefold() == category_name,
                   self.__GUILD.categories))
        # Returns a list of categories which have same name as 'category_name'
        # Should return a list containg only one category with a unique name
        # Empty list means category does not exists

        if category == []:
            # Create category if doesnot exist
            category = await self.__GUILD.create_category(category_name)
        else:
            # Get category object if exists
            category = category[0]

        ##### CHANNEL #####
        channel = list(
            filter(lambda c: c.name == channel_name, category.channels))
        # Returns a list of channels under the category which have same name as 'channel_name'
        # Should return a list containg only one channel with a unique name
        # Empty list means channel doesnot exists

        if channel == []:
            # Create a new channel under category
            channel = await category.create_text_channel(channel_name)
        else:
            # Get existing channel
            channel = channel[0]

        return category, channel

    async def new(self, category_name, channel_name):
        await self.client.wait_until_ready()
        self.__GUILD = self.client.get_guild(self.guild_id)
        category, channel = await self.__create(category_name, channel_name)
        # This is the actual class that takes care of setting and getting data from the database
        return Database(category, channel)



import json
from functools import cached_property

from .common_functions import key_check, search_key


class Database:
    def __init__(self, category_object, channel_object) -> None:
        self.__category = category_object
        self.__channel = (
            channel_object  # Channel in which all key value pairs are stored
        )

    @cached_property
    def get_channel_id(self):
        return self.__channel.id

    @cached_property
    def get_category_id(self):
        return self.__category.id

    async def set(self, key: str, value):
        key_check(str(key))
        if len(str(value)) <= 0:
            raise ValueError("value should atleast have a length of 1")

        # True/False should become 1/0
        # integers and floats should become strings
        # In the end everything is stored as a string, along with an identifier
        """ Every message should look like this
        {key:value,type:'...'}
        """

        # conversion into storable string
        if isinstance(value, bool):
            value = int(value)
        if isinstance(value, (float, int)):
            value = str(value)

        found_key, in_message, data = await search_key(str(key), self.__channel)
        if found_key:
            data[str(key)] = value
            data["type"] = value.__class__.__name__
            await in_message.edit(content=json.dumps(data))
        else:
            data = {key: value, "type": value.__class__.__name__}
            await self.__channel.send(json.dumps(data))
        return

    async def get(self, key: str):
        key_check(str(key))
        found_key, in_message, data = await search_key(key, self.__channel)
        if found_key:
            value = data[str(key)]
            value_type = data["type"]

            if value_type == "int":
                value = int(value)
            elif value_type == "float":
                value = float(value)

            if value_type == "bool":
                value = bool(int(value))
        else:
            value = None
        return value

    async def delete(self, key: str):
        key_check(str(key))
        found_key, in_message, data = await search_key(key, self.__channel)
        if found_key:
            await in_message.delete()
            return data.get(str(key))  # return the value of that key after deleting
        return  # returns None if key waws not found and nothing was deleted



# Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
from os import path

from setuptools import find_packages, setup

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# This call to setup() does all the work
setup(
    name="DiscordDatabase",
    version="0.1.3",
    description="CRUD database for discord bots, using discord text channels to store data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ankushKun/DiscordDatabase",
    author="Ankush Singh",
    author_email="ankush4singh@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    # extras_require = {
        # 'py-cord':  ["py-cord"]
    # },
    packages=find_packages(),
    include_package_data=True
)



import json
import logging
from json.decoder import JSONDecodeError

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(message)s')
ch.setFormatter(formatter)
log.addHandler(ch)

def str_is_illegal(s: str):
    # returns True if string contains any illegal characters
    legal_chars = ("_", "")
    return not all(map(lambda c: c in legal_chars or c.isalnum(), s))


def format_string(s: str):
    # replaces spaces and hyphens from string
    return s.casefold().replace("-", "").replace(" ", "")


def key_check(key: str):
    if len(key) <= 0:
        raise ValueError("key should atleast have a length of 1")

    if str_is_illegal(key):
        # 'key' should contain only alphabets and numbers
        raise ValueError("Key should contain only alphanumeric character")

    return True

async def search_key(key: str, channel):
    found_key, in_message = None, None
    async for message in channel.history(limit=None):
        cnt = message.content
        try:
            data = json.loads(str(cnt))
        except JSONDecodeError:
            logging.info(f"-----\nJSONDecodeerror: {cnt}\n-----")
            continue
        if str(key) in list(map(lambda a: str(a),list(data.keys()))):
            found_key = True
            in_message = message
            return found_key, in_message, data  # return useful data if keyis found
    return False, None, None  # return this if key is not found
